import json
import os
import importlib
from copy import deepcopy
from PySide6.QtCore import Signal, QObject, QStandardPaths
from icarus_v2.backend.event import Channel


# Responsible for loading and saving settings
class ConfigurationManager(QObject):
    FILENAME = "settings.json"
    # Signal to let subscribers know that settings[key] was changed
    settings_updated = Signal(str)

    def __init__(self) -> None:
        super().__init__()

        # Get the application configuration path
        self.config_path = os.path.join(QStandardPaths.writableLocation(QStandardPaths.AppConfigLocation), 'icarus_v2')
        os.makedirs(self.config_path, exist_ok=True)  # Create the directory if it doesn't exist

        self.filename = os.path.join(self.config_path, self.FILENAME)

        # Load application settings
        if not os.path.isfile(self.filename):
            # Load default settings from icarus_v2.resources
            with importlib.resources.path('icarus_v2.resources', 'default_settings.json') as path:
                with open(path, 'r') as file:
                    self.settings = json.load(file)
            self.save_settings()  # Save default settings to the config file
        else:
            with open(self.filename, "r") as file:
                # Dictionary
                self.settings = json.load(file)

    def get_settings(self, key):
        value = deepcopy(self.settings[key])

        # Convert to enum rather than int
        if key == 'plotting_coefficients':
            value = {Channel(int(k)): v for k, v in value.items()}

        return value

    def save_settings(self, key=None, value=None, emit=True):
        if key is not None and value is not None:
            # Convert to int rather than enum
            if key == 'plotting_coefficients':
                value = {k.value: v for k, v in value.items()}
            self.settings[key] = value

        with open(self.filename, "w") as file:
            json.dump(self.settings, file, indent=4)

        if emit and key is not None:
            self.settings_updated.emit(key)



from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QLabel, QGridLayout, QSpacerItem, QSizePolicy
import pyqtgraph as pg
from icarus_v2.backend.event import Event, Channel
from bisect import bisect_right, bisect_left
import numpy as np
from icarus_v2.gui.styled_plot_widget import StyledPlotWidget

class HistoryPlot(QWidget):
    # Dictionary of pens to plot each line
    LINE_STYLES = {
        "origin pressure": ('#FFDC00', Qt.SolidLine),  # yellow
        "sample pressure": ('#FFDC00', Qt.DashLine),  # yellow dashed
        "depress origin slope": ('#59D8E6', Qt.SolidLine),  # cyan
        "depress sample slope": ('#59D8E6', Qt.DashLine),  # cyan dashed
        "press origin slope": ('#B9121B', Qt.SolidLine),  # red
        "press sample slope": ('#B9121B', Qt.DashLine),  # red dashed
        "depress origin switch": ('#59D8E6', Qt.SolidLine),  # cyan
        "depress sample switch": ('#59D8E6', Qt.DashLine),  # cyan dashed
        "press origin switch": ('#B9121B', Qt.SolidLine),  # red
        "press sample switch": ('#B9121B', Qt.DashLine),  # red dashed
    }

    # Dictionary of lines updated on certain events
    EVENT_LINES = {
        Event.PRESSURIZE: [
            "press origin slope",
            "press sample slope",
            "press origin switch",
            "press sample switch"
        ],
        Event.DEPRESSURIZE: [
            "origin pressure",
            "sample pressure",
            "depress origin slope",
            "depress sample slope",
            "depress origin switch",
            "depress sample switch"
        ]
    }

    def __init__(self, config_manager):
        super().__init__()

        self.config_manager = config_manager
        self.coefficients = None
        self.log_coefficients = None
        self.update_settings("plotting_coefficients")
        self.config_manager.settings_updated.connect(self.update_settings)
        self.data = None
        self.initial_time = None
        self.can_plot_pressurize = None
        self.limits = None
        self.min_zoom = None

        # Dictionary of line references
        self.lines = {}
        self.hide_sample_sensor = False
        # Used to prevent plotting multiple pressurizes per depressurize which results in noisy plots
        self.can_plot_pressurize = True
        # Dictionary of mouse position labels
        self.mouse_labels = {}

        # Pressure plot
        self.pressure_plot = StyledPlotWidget(x_zoom=True)
        self.pressure_plot.set_title("Pressure")
        self.pressure_plot.set_y_label('Pressure (kBar)')
        self.pressure_plot.set_x_label('Time (s)')
        self.pressure_plot.setYRange(0, 3)
        style = self.LINE_STYLES["origin pressure"]
        pen = pg.mkPen(color=style[0], style=style[1])
        self.lines["origin pressure"] = self.pressure_plot.plot([], [], pen=pen)
        style = self.LINE_STYLES["sample pressure"]
        pen = pg.mkPen(color=style[0], style=style[1])
        self.lines["sample pressure"] = self.pressure_plot.plot([], [], pen=pen)
        self.pressure_plot.getPlotItem().getViewBox().sigStateChanged.connect(self.set_text)

        # Slope plot
        self.slope_plot = StyledPlotWidget(x_zoom=True)
        self.slope_plot.set_title("Pressure Change Slope")
        self.slope_plot.set_y_label('Slope (kBar/ms)')
        self.slope_plot.set_x_label('Time (s)')
        self.slope_plot.setYRange(-1.1, 1.1)
        style = self.LINE_STYLES["depress origin slope"]
        pen = pg.mkPen(color=style[0], style=style[1])
        self.lines["depress origin slope"] = self.slope_plot.plot([], [], pen=pen)
        style = self.LINE_STYLES["depress sample slope"]
        pen = pg.mkPen(color=style[0], style=style[1])
        self.lines["depress sample slope"] = self.slope_plot.plot([], [], pen=pen)
        style = self.LINE_STYLES["press origin slope"]
        pen = pg.mkPen(color=style[0], style=style[1])
        self.lines["press origin slope"] = self.slope_plot.plot([], [], pen=pen)
        style = self.LINE_STYLES["press sample slope"]
        pen = pg.mkPen(color=style[0], style=style[1])
        self.lines["press sample slope"] = self.slope_plot.plot([], [], pen=pen)
        # Connect the x-axis of all plots for zooming and panning
        self.slope_plot.setXLink(self.pressure_plot)

        # Switch time plot
        self.switch_time_plot = StyledPlotWidget(x_zoom=True)
        self.switch_time_plot.set_title("Switch Time")
        self.switch_time_plot.set_y_label('Time (ms)')
        self.switch_time_plot.set_x_label('Time (s)')
        self.switch_time_plot.setYRange(0, 45)
        style = self.LINE_STYLES["depress origin switch"]
        pen = pg.mkPen(color=style[0], style=style[1])
        self.lines["depress origin switch"] = self.switch_time_plot.plot([], [], pen=pen)
        style = self.LINE_STYLES["depress sample switch"]
        pen = pg.mkPen(color=style[0], style=style[1])
        self.lines["depress sample switch"] = self.switch_time_plot.plot([], [], pen=pen)
        style = self.LINE_STYLES["press origin switch"]
        pen = pg.mkPen(color=style[0], style=style[1])
        self.lines["press origin switch"] = self.switch_time_plot.plot([], [], pen=pen)
        style = self.LINE_STYLES["press sample switch"]
        pen = pg.mkPen(color=style[0], style=style[1])
        self.lines["press sample switch"] = self.switch_time_plot.plot([], [], pen=pen)
        # Connect the x-axis of all plots for zooming and panning
        self.switch_time_plot.setXLink(self.pressure_plot)

        # Statistics labels
        size = 14
        # Pressure plot
        self.last_pressure_display = QLabel("0.000")
        self.avg_pressure_display = QLabel("0.000")
        last_pressure_label = QLabel("Last:")
        avg_pressure_label = QLabel("Avg:")
        color = self.LINE_STYLES["origin pressure"][0]
        self.last_pressure_display.setStyleSheet(f"color: {color}; font-size: {size}px;")
        self.avg_pressure_display.setStyleSheet(f"color: {color}; font-size: {size}px;")
        last_pressure_label.setStyleSheet(f"color: {color}; font-size: {size}px;")
        avg_pressure_label.setStyleSheet(f"color: {color}; font-size: {size}px;")
        # Mouse label
        self.mouse_labels[self.pressure_plot] = QLabel("")
        self.mouse_labels[self.pressure_plot].setStyleSheet(f"font-size: {size}px;")

        # Limit rate of mouseMoved signal to 60 Hz
        def pressure_func(x):
            self.mouse_moved(self.pressure_plot, x)

        self.pressure_proxy = pg.SignalProxy(self.pressure_plot.scene().sigMouseMoved, rateLimit=120,
                                             slot=pressure_func)
        spacer = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Expanding)
        pressure_labels = QGridLayout()
        pressure_labels.setContentsMargins(0, 35, 5, 45)
        pressure_labels.addWidget(last_pressure_label, 0, 1)
        pressure_labels.addWidget(avg_pressure_label, 1, 1)
        pressure_labels.addWidget(self.last_pressure_display, 0, 2)
        pressure_labels.addWidget(self.avg_pressure_display, 1, 2)
        pressure_labels.addItem(spacer, 2, 0)
        pressure_labels.addWidget(self.mouse_labels[self.pressure_plot], 3, 0, 1, 3, Qt.AlignRight | Qt.AlignBottom)
        # Slope plot
        self.last_press_slope_display = QLabel("0.00")
        self.avg_press_slope_display = QLabel("0.00")
        last_press_slope_label = QLabel("Last:")
        avg_press_slope_label = QLabel("Avg:")
        color = self.LINE_STYLES["press origin slope"][0]
        self.last_press_slope_display.setStyleSheet(f"color: {color}; font-size: {size}px;")
        self.avg_press_slope_display.setStyleSheet(f"color: {color}; font-size: {size}px;")
        last_press_slope_label.setStyleSheet(f"color: {color}; font-size: {size}px;")
        avg_press_slope_label.setStyleSheet(f"color: {color}; font-size: {size}px;")
        self.last_depress_slope_display = QLabel("0.00")
        self.avg_depress_slope_display = QLabel("0.00")
        last_depress_slope_label = QLabel("Last:")
        avg_depress_slope_label = QLabel("Avg:")
        color = self.LINE_STYLES["depress origin slope"][0]
        self.last_depress_slope_display.setStyleSheet(f"color: {color}; font-size: {size}px;")
        self.avg_depress_slope_display.setStyleSheet(f"color: {color}; font-size: {size}px;")
        last_depress_slope_label.setStyleSheet(f"color: {color}; font-size: {size}px;")
        avg_depress_slope_label.setStyleSheet(f"color: {color}; font-size: {size}px;")
        # Mouse label
        self.mouse_labels[self.slope_plot] = QLabel("")
        self.mouse_labels[self.slope_plot].setStyleSheet(f"font-size: {size}px;")

        # Limit rate of mouseMoved signal to 60 Hz

        def slope_func(x):
            self.mouse_moved(self.slope_plot, x)

        self.slope_proxy = pg.SignalProxy(self.slope_plot.scene().sigMouseMoved, rateLimit=120, slot=slope_func)
        spacer = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Expanding)
        slope_labels = QGridLayout()
        slope_labels.setContentsMargins(0, 35, 5, 45)
        slope_labels.addWidget(last_press_slope_label, 0, 1)
        slope_labels.addWidget(avg_press_slope_label, 1, 1)
        slope_labels.addWidget(self.last_press_slope_display, 0, 2)
        slope_labels.addWidget(self.avg_press_slope_display, 1, 2)
        slope_labels.addWidget(last_depress_slope_label, 2, 1)
        slope_labels.addWidget(avg_depress_slope_label, 3, 1)
        slope_labels.addWidget(self.last_depress_slope_display, 2, 2)
        slope_labels.addWidget(self.avg_depress_slope_display, 3, 2)
        slope_labels.addItem(spacer, 4, 0)
        slope_labels.addWidget(self.mouse_labels[self.slope_plot], 5, 0, 1, 3, Qt.AlignRight | Qt.AlignBottom)
        # Switch Time Plot
        self.last_press_switch_display = QLabel("0.00")
        self.avg_press_switch_display = QLabel("0.00")
        last_press_switch_label = QLabel("Last:")
        avg_press_switch_label = QLabel("Avg:")
        color = self.LINE_STYLES["press origin switch"][0]
        self.last_press_switch_display.setStyleSheet(f"color: {color}; font-size: {size}px;")
        self.avg_press_switch_display.setStyleSheet(f"color: {color}; font-size: {size}px;")
        last_press_switch_label.setStyleSheet(f"color: {color}; font-size: {size}px;")
        avg_press_switch_label.setStyleSheet(f"color: {color}; font-size: {size}px;")
        self.last_depress_switch_display = QLabel("0.00")
        self.avg_depress_switch_display = QLabel("0.00")
        last_depress_switch_label = QLabel("Last:")
        avg_depress_switch_label = QLabel("Avg:")
        color = self.LINE_STYLES["depress origin switch"][0]
        self.last_depress_switch_display.setStyleSheet(f"color: {color}; font-size: {size}px;")
        self.avg_depress_switch_display.setStyleSheet(f"color: {color}; font-size: {size}px;")
        last_depress_switch_label.setStyleSheet(f"color: {color}; font-size: {size}px;")
        avg_depress_switch_label.setStyleSheet(f"color: {color}; font-size: {size}px;")
        # Mouse label
        self.mouse_labels[self.switch_time_plot] = QLabel("")
        self.mouse_labels[self.switch_time_plot].setStyleSheet(f"font-size: {size}px;")

        # Limit rate of mouseMoved signal to 60 Hz
        def switch_func(x):
            self.mouse_moved(self.switch_time_plot, x)

        self.switch_proxy = pg.SignalProxy(self.switch_time_plot.scene().sigMouseMoved, rateLimit=120, slot=switch_func)
        spacer = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Expanding)
        switch_labels = QGridLayout()
        switch_labels.setContentsMargins(0, 35, 5, 45)
        switch_labels.addWidget(last_press_switch_label, 0, 1)
        switch_labels.addWidget(avg_press_switch_label, 1, 1)
        switch_labels.addWidget(self.last_press_switch_display, 0, 2)
        switch_labels.addWidget(self.avg_press_switch_display, 1, 2)
        switch_labels.addWidget(last_depress_switch_label, 2, 1)
        switch_labels.addWidget(avg_depress_switch_label, 3, 1)
        switch_labels.addWidget(self.last_depress_switch_display, 2, 2)
        switch_labels.addWidget(self.avg_depress_switch_display, 3, 2)
        switch_labels.addItem(spacer, 4, 0)
        switch_labels.addWidget(self.mouse_labels[self.switch_time_plot], 5, 0, 1, 3, Qt.AlignRight | Qt.AlignBottom)

        # Lines for displaying log view time
        self.press_time_press = None
        self.press_time_slope = None
        self.press_time_switch = None
        self.depress_time_press = None
        self.depress_time_slope = None
        self.depress_time_switch = None

        # Set layout
        self.layout = QGridLayout()
        self.layout.addWidget(self.pressure_plot, 0, 0)
        self.layout.addLayout(pressure_labels, 0, 0, Qt.AlignRight | Qt.AlignTop)
        self.layout.addWidget(self.slope_plot, 1, 0)
        self.layout.addLayout(slope_labels, 1, 0, Qt.AlignRight | Qt.AlignTop)
        self.layout.addWidget(self.switch_time_plot, 2, 0)
        self.layout.addLayout(switch_labels, 2, 0, Qt.AlignRight | Qt.AlignTop)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        self.reset_history()

    def reset_history(self):
        self.data = {
            "time": {
                Event.DEPRESSURIZE: [],
                Event.PRESSURIZE: []
            },
            "origin pressure": [],
            "sample pressure": [],
            "depress origin slope": [],
            "depress sample slope": [],
            "press origin slope": [],
            "press sample slope": [],
            "depress origin switch": [],
            "depress sample switch": [],
            "press origin switch": [],
            "press sample switch": [],
        }
        self.initial_time = None
        self.can_plot_pressurize = True

        for line in self.lines.values():
            line.setData([], [])

        self.limits = (0, 1)
        self.min_zoom = 1
        self.reset_limits()

        self.last_pressure_display.setText("0.000")
        self.avg_pressure_display.setText("0.000")
        self.last_press_slope_display.setText("0.00")
        self.avg_press_slope_display.setText("0.00")
        self.last_depress_slope_display.setText("0.00")
        self.avg_depress_slope_display.setText("0.00")
        self.last_press_switch_display.setText("0.00")
        self.avg_press_switch_display.setText("0.00")
        self.last_depress_switch_display.setText("0.00")
        self.avg_depress_switch_display.setText("0.00")

        # Lines for displaying log view time
        self.pressure_plot.removeItem(self.press_time_press)
        self.slope_plot.removeItem(self.press_time_slope)
        self.switch_time_plot.removeItem(self.press_time_switch)
        self.pressure_plot.removeItem(self.depress_time_press)
        self.slope_plot.removeItem(self.depress_time_slope)
        self.switch_time_plot.removeItem(self.depress_time_switch)

        self.pressure_plot.setXRange(self.limits[0], self.limits[1])

    def add_event(self, event):
        # plot only one pressurize per pressurize. plots are noisy with multiple pressurizes.
        if event.event_type == Event.PRESSURIZE:
            if not self.can_plot_pressurize:
                return
            self.can_plot_pressurize = False
        elif event.event_type == Event.DEPRESSURIZE:
            self.can_plot_pressurize = True

        # Check if currently fully zoomed out
        current_x_range = self.pressure_plot.viewRange()[0]
        low_diff = current_x_range[0] - self.limits[0]
        hi_diff = self.limits[1] - current_x_range[1]
        currently_zoomed_out = low_diff + hi_diff < 1

        self.process_data(event)
        for update in self.EVENT_LINES[event.event_type]:
            self.lines[update].setData(self.data["time"][event.event_type], self.data[update])

        # Update limits to fit new point
        max_view = max(event.event_time - self.initial_time, self.pressure_plot.viewRange()[0][1])
        self.limits = (0, max_view)
        if self.min_zoom < 10 <= max_view:
            self.min_zoom = 10
        self.reset_limits()
        # Update view range iff was already zoomed out
        if currently_zoomed_out:
            # Only need to do pressure plot since plot x ranges are linked
            self.pressure_plot.setXRange(self.limits[0], self.limits[1])

    def reset_limits(self):
        self.pressure_plot.setLimits(xMin=self.limits[0], xMax=self.limits[1], minXRange=self.min_zoom)
        self.slope_plot.setLimits(xMin=self.limits[0], xMax=self.limits[1], minXRange=self.min_zoom)
        self.switch_time_plot.setLimits(xMin=self.limits[0], xMax=self.limits[1], minXRange=self.min_zoom)

    # Assumes list is sorted by time
    def load_event_list(self, event_list):
        if len(event_list) == 0:
            return

        for event in event_list:
            if event.event_type == Event.DEPRESSURIZE:
                self.can_plot_pressurize = True
                self.process_data(event)
            if event.event_type == Event.PRESSURIZE and self.can_plot_pressurize:
                self.can_plot_pressurize = False
                self.process_data(event)

        for update in self.EVENT_LINES[Event.DEPRESSURIZE]:
            self.lines[update].setData(self.data["time"][Event.DEPRESSURIZE], self.data[update])
        for update in self.EVENT_LINES[Event.PRESSURIZE]:
            self.lines[update].setData(self.data["time"][Event.PRESSURIZE], self.data[update])

        max_view = max(event_list[-1].event_time - self.initial_time, self.pressure_plot.viewRange()[0][1])
        self.limits = (0, max_view)
        if self.min_zoom < 10 <= max_view:
            self.min_zoom = 10
        self.reset_limits()
        # Only need to do pressure plot since plot x ranges are linked
        self.pressure_plot.setXRange(self.limits[0], self.limits[1])
        self.set_text()

    def process_data(self, event):
        # Define start time of plots
        if self.initial_time is None:
            self.initial_time = event.event_time

        # Update data
        self.data["time"][event.event_type].append(event.event_time - self.initial_time)
        coefficients = self.coefficients if self.log_coefficients is None else self.log_coefficients
        for update in self.EVENT_LINES[event.event_type]:
            # Add event to data dict
            self.data[update].append(event.get_event_info(update) * coefficients[update])

    # Sets all text to the data in view
    def set_text(self):
        view_range = self.pressure_plot.viewRange()[0]
        press_range = (
            bisect_left(self.data["time"][Event.PRESSURIZE], view_range[0]),
            bisect_right(self.data["time"][Event.PRESSURIZE], view_range[1]) - 1
        )
        depress_range = (
            bisect_left(self.data["time"][Event.DEPRESSURIZE], view_range[0]),
            bisect_right(self.data["time"][Event.DEPRESSURIZE], view_range[1]) - 1
        )

        if depress_range[1] - depress_range[0] > 0:
            self.last_pressure_display.setText(f"{self.data['origin pressure'][depress_range[1]]:.3f}")
            self.avg_pressure_display.setText(
                f"{np.mean(self.data['origin pressure'][depress_range[0]:depress_range[1]]):.3f}"
            )
            self.last_depress_slope_display.setText(f"{self.data['depress origin slope'][depress_range[1]]:.2f}")
            self.avg_depress_slope_display.setText(
                f"{np.mean(self.data['depress origin slope'][depress_range[0]:depress_range[1]]):.2f}"
            )
            self.last_depress_switch_display.setText(f"{self.data['depress origin switch'][depress_range[1]]:.2f}")
            self.avg_depress_switch_display.setText(
                f"{np.mean(self.data['depress origin switch'][depress_range[0]:depress_range[1]]):.2f}"
            )
        if press_range[1] - press_range[0] > 0:
            self.last_press_slope_display.setText(f"{self.data['press origin slope'][press_range[1]]:.2f}")
            self.avg_press_slope_display.setText(
                f"{np.mean(self.data['press origin slope'][press_range[0]:press_range[1]]):.2f}"
            )
            self.last_press_switch_display.setText(f"{self.data['press origin switch'][press_range[1]]:.2f}")
            self.avg_press_switch_display.setText(
                f"{np.mean(self.data['press origin switch'][press_range[0]:press_range[1]]):.2f}"
            )

    def update_settings(self, key):
        if key == "plotting_coefficients":
            plotting_coefficients = self.config_manager.get_settings(key)
            self.coefficients = self.define_coefficients(plotting_coefficients)

    def render_pressurize_time(self, event):
        # Remove old time
        self.pressure_plot.removeItem(self.press_time_press)
        self.slope_plot.removeItem(self.press_time_slope)
        self.switch_time_plot.removeItem(self.press_time_switch)

        if event is not None:
            time = event.event_time - self.initial_time
            style = self.LINE_STYLES["press origin slope"]
            pen = pg.mkPen(color=style[0], style=style[1])
            self.press_time_press = pg.InfiniteLine(pos=time, angle=90, movable=False, pen=pen)
            self.press_time_slope = pg.InfiniteLine(pos=time, angle=90, movable=False, pen=pen)
            self.press_time_switch = pg.InfiniteLine(pos=time, angle=90, movable=False, pen=pen)
            self.pressure_plot.addItem(self.press_time_press)
            self.slope_plot.addItem(self.press_time_slope)
            self.switch_time_plot.addItem(self.press_time_switch)

    def render_depressurize_time(self, event):
        # Remove old time
        self.pressure_plot.removeItem(self.depress_time_press)
        self.slope_plot.removeItem(self.depress_time_slope)
        self.switch_time_plot.removeItem(self.depress_time_switch)

        if event is not None:
            time = event.event_time - self.initial_time
            style = self.LINE_STYLES["depress origin slope"]
            pen = pg.mkPen(color=style[0], style=style[1])
            self.depress_time_press = pg.InfiniteLine(pos=time, angle=90, movable=False, pen=pen)
            self.depress_time_slope = pg.InfiniteLine(pos=time, angle=90, movable=False, pen=pen)
            self.depress_time_switch = pg.InfiniteLine(pos=time, angle=90, movable=False, pen=pen)
            self.pressure_plot.addItem(self.depress_time_press)
            self.slope_plot.addItem(self.depress_time_slope)
            self.switch_time_plot.addItem(self.depress_time_switch)

    def mouse_moved(self, plot, event):
        mouse_point = plot.getViewBox().mapSceneToView(event[0])
        view_range = plot.getViewBox().viewRange()

        # Check if the mouse point is within the view range
        if (view_range[0][0] <= mouse_point.x() <= 0.98 * view_range[0][1] and
                view_range[1][0] <= mouse_point.y() <= view_range[1][1]):
            self.mouse_labels[plot].setText(f"{mouse_point.x():.2f}, {mouse_point.y():.2f}")
        else:
            self.mouse_labels[plot].setText("")

    def set_sample_sensor(self, connected):
        if not connected and not self.hide_sample_sensor:
            self.pressure_plot.removeItem(self.lines["sample pressure"])
            self.slope_plot.removeItem(self.lines["press sample slope"])
            self.slope_plot.removeItem(self.lines["depress sample slope"])
            self.switch_time_plot.removeItem(self.lines["press sample switch"])
            self.switch_time_plot.removeItem(self.lines["depress sample switch"])
            self.hide_sample_sensor = True

        elif connected and self.hide_sample_sensor:
            self.pressure_plot.addItem(self.lines["sample pressure"])
            self.slope_plot.addItem(self.lines["press sample slope"])
            self.slope_plot.addItem(self.lines["depress sample slope"])
            self.switch_time_plot.addItem(self.lines["press sample switch"])
            self.switch_time_plot.addItem(self.lines["depress sample switch"])
            self.hide_sample_sensor = False

    def set_log_coefficients(self, coefficients):
        self.log_coefficients = self.define_coefficients(coefficients)

    # Dictionary of coefficient to apply when plotting each channel
    @staticmethod
    def define_coefficients(plotting_coefficients):
        if plotting_coefficients is None:
            return None
        # Dictionary of coefficient to apply when plotting each channel
        coefficients = {
            "origin pressure": plotting_coefficients[Channel.HI_PRE_ORIG],
            "sample pressure": plotting_coefficients[Channel.HI_PRE_SAMPLE],
            "depress origin slope": plotting_coefficients[Channel.HI_PRE_ORIG] * 4,
            "depress sample slope": plotting_coefficients[Channel.HI_PRE_SAMPLE] * 4,
            "press origin slope": plotting_coefficients[Channel.HI_PRE_ORIG] * 4,
            "press sample slope": plotting_coefficients[Channel.HI_PRE_SAMPLE] * 4,
            "depress origin switch": 0.25,
            "depress sample switch": 0.25,
            "press origin switch": 0.25,  # ms per index
            "press sample switch": 0.25,  # ms per index
        }
        return coefficients



from array import array
from threading import Lock
import lzma
import pickle
import numpy as np
import csv


# Fake device to load example data files
# Used only for testing.
class RawLogReader:
    def __init__(self, filename) -> None:
        self.sample_rate = 4000
        self.points_to_read = 64
        self.channels_to_read = 8
        self.initial_time = None
        self.current_dio = None
        self.acquiring = None
        self.bytes_to_read = self.channels_to_read * 2 * self.points_to_read

        # Used to tell how long to wait on reads
        self.read_count = 0
        # File
        self.file = lzma.open(filename, "rb")

    def read_data(self):
        self.read_count += 1

        try:
            # Deserialize each object from the file
            data = pickle.load(self.file)
            # Append the array if it matches the expected structure
            if not isinstance(data, array):
                raise TypeError("Error parsing example data file: Data is not an array.")
            return data
        except EOFError:
            raise RuntimeError("End of file reached.")

    def process_data(self, data):
        data_shape = (64, 8)

        int_array = np.frombuffer(data, dtype=np.int16)
        reshaped_array = np.reshape(int_array, data_shape)
        reshaped_array[:,-1] = reshaped_array[:,-1] >> 8    # Digital is only in the 1st byte

        return reshaped_array

    def close_device(self):
        self.file.close()


def read_and_write_to_csv(raw_log_reader, output_filename):
    with open(output_filename, 'w', newline='') as csvfile:
        csv_writer = csv.writer(csvfile)

        # Write header
        header = [f'Channel_{i}' for i in range(raw_log_reader.channels_to_read - 1)] + ['Digital']
        csv_writer.writerow(header)

        try:
            while True:
                raw_data = raw_log_reader.read_data()
                processed_data = raw_log_reader.process_data(raw_data)

                for row in processed_data:
                    csv_writer.writerow(row)

        except RuntimeError as e:
            if str(e) == "End of file reached.":
                print("Finished reading the file.")
            else:
                print(f"An error occurred: {e}")
        finally:
            raw_log_reader.close_device()


# Usage
input_file = "logs/example/log_2024-07-19_18-19-00.xz"
output_file = "output.csv"

reader = RawLogReader(input_file)
read_and_write_to_csv(reader, output_file)


from PySide6.QtCore import Qt
import pyqtgraph as pg
import numpy as np
from icarus_v2.backend.event import Event, Channel, get_channel
from PySide6.QtWidgets import QLabel, QGridLayout, QWidget, QSpacerItem, QSizePolicy
from icarus_v2.gui.styled_plot_widget import StyledPlotWidget


class EventPlot(QWidget):
    # Dictionary of color to plot each channel
    LINE_STYLES = {
        Channel.TARGET: ('#45BF55', Qt.SolidLine),          # light green
        Channel.DEPRE_LOW: ('#AB47BC', Qt.SolidLine),       # magenta
        Channel.DEPRE_UP: ('#004B8D', Qt.SolidLine),        # blue
        Channel.PRE_LOW: ('#AB47BC', Qt.SolidLine),         # magenta
        Channel.PRE_UP: ('#004B8D', Qt.SolidLine),          # blue
        Channel.HI_PRE_ORIG: ('#FFDC00', Qt.SolidLine),     # yellow
        Channel.HI_PRE_SAMPLE: ('#FFDC00', Qt.DashLine),    # yellow dashed
        Channel.DEPRE_VALVE: ('#59D8E6', Qt.SolidLine),     # cyan
        Channel.PRE_VALVE: ('#B9121B', Qt.SolidLine),       # red
    }

    def __init__(self, event_type, config_manager, parent=None):
        super().__init__(parent=parent)

        self.event_type = event_type
        self.config_manager = config_manager
        self.config_manager.settings_updated.connect(self.update_settings)
        self.coefficients = config_manager.get_settings("plotting_coefficients")
        self.log_coefficients = None
        self.hide_valve_setting = config_manager.get_settings("hide_valve_sensors")
        self.hide_sample_sensor = False
        display_channels = [
            Channel.TARGET, Channel.HI_PRE_SAMPLE, Channel.HI_PRE_ORIG, Channel.DEPRE_VALVE, Channel.PRE_VALVE
        ]

        if event_type == Event.PRESSURIZE:
            display_channels += [Channel.PRE_LOW, Channel.PRE_UP]
            self.x_unit = 'ms'
            title = "Pressurize"
        elif event_type == Event.DEPRESSURIZE:
            display_channels += [Channel.DEPRE_LOW, Channel.DEPRE_UP]
            self.x_unit = 'ms'
            title = "Depressurize"
        else:
            self.x_unit = 's'
            title = "Period"

        self.plot = StyledPlotWidget()
        self.plot.setYRange(0, 3)
        self.plot.set_title(title)
        self.plot.set_y_label('Pressure (kbar)')
        self.plot.set_x_label(f'Time ({self.x_unit})')
        self.width_display = None
        self.period_display = None
        self.delay_display = None
        self.mouse_label = None
        self.proxy = None

        # Create a dictionary of lines for each channel listed in display_channels
        self.line_references = {}
        self.line_currently_shown = {}
        for channel in display_channels:
            style = self.LINE_STYLES[channel]
            pen = pg.mkPen(color=style[0], style=style[1])
            self.line_references[channel] = self.plot.plot([], [], pen=pen)
            self.line_currently_shown[channel] = True

        labels = self.init_labels()
        self.hide_valve_sensors(self.hide_valve_setting)

        # Set layout
        layout = QGridLayout()
        layout.addWidget(self.plot, 0, 0)
        layout.addLayout(labels, 0, 0, Qt.AlignRight | Qt.AlignTop)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def init_labels(self):
        layout = QGridLayout()
        press_color = self.LINE_STYLES[Channel.PRE_VALVE][0]
        depress_color = self.LINE_STYLES[Channel.DEPRE_VALVE][0]
        size = 14
        if self.event_type == Event.PRESSURIZE or self.event_type == Event.DEPRESSURIZE:
            color = press_color if self.event_type == Event.PRESSURIZE else depress_color
            self.width_display = QLabel("0.00")
            label = QLabel("Valve Open (ms):")
            self.width_display.setStyleSheet(f"color: {color}; font-size: {size}px;")
            label.setStyleSheet(f"color: {color}; font-size: {size}px;")
            layout.addWidget(label, 0, 0)
            layout.addWidget(self.width_display, 0, 1)

        elif self.event_type == Event.PERIOD:
            self.period_display = QLabel("0")
            self.delay_display = QLabel("0")
            period_label = QLabel("Period (s):")
            delay_label = QLabel("Delay (s):")
            self.period_display.setStyleSheet(f"color: {depress_color}; font-size: {size}px;")
            self.delay_display.setStyleSheet(f"color: {press_color}; font-size: {size}px;")
            period_label.setStyleSheet(f"color: {depress_color}; font-size: {size}px;")
            delay_label.setStyleSheet(f"color: {press_color}; font-size: {size}px;")
            layout.addWidget(period_label, 0, 0)
            layout.addWidget(delay_label, 1, 0)
            layout.addWidget(self.period_display, 0, 1)
            layout.addWidget(self.delay_display, 1, 1)

        # Mouse position label
        spacer = QSpacerItem(0, 0, QSizePolicy.Preferred, QSizePolicy.Expanding)
        self.mouse_label = QLabel("")
        self.mouse_label.setStyleSheet(f"font-size: {size}px;")
        # Limit rate of mouseMoved signal to 60 Hz
        self.proxy = pg.SignalProxy(self.plot.scene().sigMouseMoved, rateLimit=60, slot=self.mouse_moved)
        layout.addItem(spacer, 2, 0)
        layout.addWidget(self.mouse_label, 3, 0, 1, 2, Qt.AlignRight | Qt.AlignBottom)

        layout.setContentsMargins(0, 35, 5, 45)
        return layout

    def update_data(self, event):
        if event is None:
            self.reset_history()
            return

        data = event.data
        # Calculate times based on event.step_time and event_index
        time_before_event = - event.event_index * event.step_time
        time_after_event = (len(event.data) - event.event_index - 1) * event.step_time
        times = np.linspace(time_before_event, time_after_event, len(data))
        if self.x_unit == "s":
            times /= 1000

        # update data for each line
        coefficients = self.coefficients if self.log_coefficients is None else self.log_coefficients
        for channel, line_reference in self.line_references.items():
            coefficient = coefficients[channel]
            line_reference.setData(times, get_channel(data, channel) * coefficient)

        # Update timings labels
        if event.event_type == Event.PRESSURIZE or event.event_type == Event.DEPRESSURIZE:
            width = event.get_valve_open_time() / 4
            self.width_display.setText(f"{width:.2f}")

        elif event.event_type == Event.PERIOD:
            period_width = event.get_period_width() / 4000
            delay_width = event.get_delay_width() / 4000
            self.period_display.setText(f"{period_width:.2f}")
            self.delay_display.setText(f"{delay_width:.2f}")

    def update_settings(self, key):
        if key == "plotting_coefficients":
            self.coefficients = self.config_manager.get_settings(key)
        elif key == "hide_valve_sensors":
            self.hide_valve_setting = self.config_manager.get_settings(key)
            self.hide_valve_sensors(self.hide_valve_setting)

    def reset_history(self):
        for line_reference in self.line_references.values():
            line_reference.setData([], [])

    def mouse_moved(self, event):
        mouse_point = self.plot.getViewBox().mapSceneToView(event[0])
        view_range = self.plot.getViewBox().viewRange()

        # Check if the mouse point is within the view range
        if (view_range[0][0] <= mouse_point.x() <= 0.98 * view_range[0][1] and
                view_range[1][0] <= mouse_point.y() <= view_range[1][1]):
            self.mouse_label.setText(f"{mouse_point.x():.2f}, {mouse_point.y():.2f}")
        else:
            self.mouse_label.setText("")

    def hide_valve_sensors(self, hide_valve_sensors):
        if self.event_type == Event.PRESSURIZE:
            valve_sensors = [Channel.PRE_LOW, Channel.PRE_UP]
        elif self.event_type == Event.DEPRESSURIZE:
            valve_sensors = [Channel.DEPRE_LOW, Channel.DEPRE_UP]
        else:
            valve_sensors = []

        if hide_valve_sensors:
            for channel in valve_sensors:
                if self.line_currently_shown[channel]:
                    self.plot.removeItem(self.line_references[channel])
                    self.line_currently_shown[channel] = False
        else:
            for channel in valve_sensors:
                if not self.line_currently_shown[channel]:
                    self.plot.addItem(self.line_references[channel])
                    self.line_currently_shown[channel] = True

    def set_sample_sensor(self, connected):
        if not connected and not self.hide_sample_sensor:
            self.plot.removeItem(self.line_references[Channel.HI_PRE_SAMPLE])
            self.hide_sample_sensor = True
        elif connected and self.hide_sample_sensor:
            self.plot.addItem(self.line_references[Channel.HI_PRE_SAMPLE])
            self.hide_sample_sensor = False

    def set_log_coefficients(self, coefficients):
        self.log_coefficients = coefficients



import importlib
import subprocess


# Adds permissions for the user to access the USB device
def setup_udev_rules():
    with importlib.resources.path('icarus_v2.resources.setup', 'install_udev_rules.sh') as script_path:
        try:
            # Run the script with 'pkexec'
            command = ['pkexec', str(script_path)]  # Convert PosixPath to string
            result = subprocess.run(command, stderr=subprocess.PIPE)
            if result.returncode != 0:
                raise RuntimeError(f"Command failed: {result.stderr.decode().strip()}")
        except Exception as e:
            raise RuntimeError(f"Command execution failed: {str(e)}")



import lzma
import pickle
import os
from datetime import datetime
from PySide6.QtCore import QStandardPaths


# Logs files to logs/temp or logs/experiment depending on bit 4.
# Files are deleted if no events are logged.
class Logger:
    def __init__(self, config_manager=None) -> None:
        self.config_manager = config_manager
        self.file = None
        self.filename = None
        self.current_path = None
        self.event_count = None

    def new_log_file(self, temporary=True, raw=False):
        self.close()

        self.current_path = temporary
        base_dir = os.path.join(QStandardPaths.writableLocation(QStandardPaths.AppConfigLocation), 'icarus_v2', 'logs')
        if raw:
            log_path = os.path.join(base_dir, "example")
        elif temporary:
            log_path = os.path.join(base_dir, 'temp')
        else:
            log_path = os.path.join(base_dir, "experiment")

        if not os.path.exists(log_path):
            os.makedirs(log_path)

        current_datetime = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        self.filename = f"{log_path}/log_{current_datetime}.xz"
        self.file = lzma.open(self.filename, "ab")  # Opening file in append binary mode with LZMA compression
        self.event_count = 0

    def log_event(self, event):
        event_dict = {
            'event_type': event.event_type,
            'data': event.data,
            'event_time': event.event_time,
            'event_index': event.event_index,
            'step_time': event.step_time
        }   
        self.log_raw(event_dict)

    def log_raw(self, data):
        self.event_count += 1
        pickle.dump(data, self.file, protocol=pickle.HIGHEST_PROTOCOL)

    def flush(self):
        self.file.close()
        self.file = lzma.open(self.filename, "ab")

    def close(self):
        if self.file is not None:
            if self.config_manager is not None:
                settings = {"plotting_coefficients": self.config_manager.get_settings("plotting_coefficients")}
                pickle.dump(settings, self.file, protocol=pickle.HIGHEST_PROTOCOL)
            self.file.close()
            self.file = None
            if self.event_count == 0:
                try:
                    os.remove(self.filename)
                except FileNotFoundError:
                    pass

            self.filename = None



from pyqtgraph import PlotWidget
from icarus_v2.qdarktheme.load_style import THEME_COLOR_VALUES


class StyledPlotWidget(PlotWidget):
    def __init__(self, x_zoom=False):
        theme = 'dark'
        background = THEME_COLOR_VALUES[theme]['background']['base']
        self.text_color = THEME_COLOR_VALUES[theme]['foreground']['base']
        PlotWidget.__init__(self, background=background)

        self.showGrid(x=True, y=True)
        self.setMouseEnabled(x=x_zoom, y=False)  # Prevent zooming
        self.hideButtons()  # Remove autoScale button

    def set_title(self, title):
        self.setTitle(title, color=self.text_color, size="17pt")

    def set_y_label(self, label):
        self.setLabel('left', label, **{'color': self.text_color})

    def set_x_label(self, label):
        self.setLabel('bottom', label, **{'color': self.text_color})
